﻿--[[

	LibRecursiveAnim - World of Warcraft addon that allows animate frame and his children as single object

	Copyright (C) 2010 Ruslan Burykin <Heckffy@gmail.com>

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 2
	of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

]]

--[[
	CreateRecursiveAnimationGroup - AnimationGroup constructor, similar to Frame:CreateAnimationGroup(), as well
		it allows animate frame and its children as single object
		
	Usage:
		libAnim = LibStub:GetLibrary("LibRecursiveAnim-1.0")
	
		recursiveAnimationGroup = libAnim:CreateRecursiveAnimationGroup(frame, name, inheritsFrom)
		
			frame - frame that should be animated (Frame)
			name  - optional, global identifier (string)
			inheritsFrom - optional, temlate for animationGroup (string)
			
			recursiveAnimationGroup - object similar to AnimationGroup, except controlling not only given frame,
				but all his children also
				
		recursiveAnimationGroup:Refresh()
		
			this method should be called whenever frame resized, moved or frame structure is changed
			
		recursiveAnimation = recursiveAnimationGroup:CreateAnimation(animationType, name, inheritsFrom)
		
			animationType - animation type: "Alpha", "Scale", "Translation", "Rotation" or "Path" (string)
			name         - optional, global identifier (string)
			inheritsFrom - optional, temlate for animation (string)
			
			recursiveAnimation - object similar to Animation
			
		recursiveControlPoint = recursiveAnimation:CreateControlPoint(name, inheritsFrom, order)
			
			only Path animation has this method
			
			name         - optional, global identifier (string)
			inheritsFrom - optional, temlate for controlPoint (string)
			order        - optional, control point order, I strongly recommend set it here instead of using 
				:SetOrder(),	as it can crash client in some cases
				
			recursiveControlPoint - object similar to ControlPoint
			
		recursiveAnimation:SetParent()
		recursiveControlPoint:SetParent()
		
			not implemented
			
		type = recursiveAnimationGroup:GetObjectType()
		type = recursiveAnimation:GetObjectType()
		type = recursiveControlPoint:GetObjectType()
			
			type - "RecursiveAnimationGroup", "RecursiveAnimation" or "RecursiveControlPoint" respectively
			
		isObjectType = recursiveAnimationGroup:IsObjectType(type)
		isObjectType = recursiveAnimation:IsObjectType(type)
		isObjectType = recursiveControlPoint:IsObjectType(type)
		
			type - expected type (string)
			
			isObjectType - true, if type equals "RecursiveAnimationGroup", "RecursiveAnimation" or 
				"RecursiveControlPoint" respectively, false otherwise
				
		recursiveAnimationGroup:SetScript(handler, function, firstOnly)
		recursiveAnimation:SetScript(handler, function, firstOnly)
			
			handler   - handler type (string)
			function  - handler function (function)
			firstOnly - optional, if this flag set script will be set for root frame only, and will be triggered 
				only once per event, if it's not set script will be set for frame and each his child, by default 
				it's false (boolean)
				
		function = recursiveAnimationGroup:GetScript(handler, firstOnly)
		function = recursiveAnimation:GetScript(handler, firstOnly)
			
			handler   - handler type (string)
			firstOnly - optional, if this flag set method will return handler function of root frame, otherwise 
				it will return handler function of child frame, or, if there aren't any child frames - function 
				of root frame, by default it's false (boolean)
				
			function - handler function
				
		hasScript = recursiveAnimationGroup:HasScript(handler, checkFirstOnly)
		hasScript = recursiveAnimation:HasScript(handler, checkFirstOnly)
			
			handler - handler type (string)
			checkFirstOnly - optional, if this flag set method will return check handler function of root frame only,
				otherwise it will check child frame, or, if there aren't child frames it will check root frame, by 
				default it's false (boolean)
				
			hasScript - true if there are handler function, false otherwise
				
		recursiveAnimationGroup:HookScript(handler, function, firstOnly)
		recursiveAnimation:HookScript(handler, function, firstOnly)
			
			handler   - handler type (string)
			function  - handler function (function)
			firstOnly - optional, if this flag set script will be hooked for root frame only, otherwise it will 
				be hooked for frame and each his child, by default it's false (boolean)

	Example:
		libAnim = LibStub:GetLibrary("LibRecursiveAnim-1.0");
		
		local MyFrame1 = CreateFrame("Frame", nil, UIParent);
		MyFrame1:SetHeight(100);
		MyFrame1:SetWidth(100);
		MyFrame1:SetPoint("CENTER");
		local Texture1 = MyFrame1:CreateTexture();
		Texture1:SetAllPoints();
		Texture1:SetTexture(0, 0, 0);

		local animationGroup = libAnim:CreateRecursiveAnimationGroup(MyFrame1);
		local animation = animationGroup:CreateAnimation("Rotation");
		animation:SetDegrees(180);
		animation:SetDuration(10);
		
		local MyFrame2 = CreateFrame("Frame", nil, MyFrame1);
		MyFrame2:SetHeight(100);
		MyFrame2:SetWidth(100);
		MyFrame2:SetPoint("BOTTOMLEFT", MyFrame1, "TOPRIGHT");
		local Texture2 = MyFrame2:CreateTexture();
		Texture2:SetAllPoints();
		Texture2:SetTexture(0, 0, 0);
		
		animationGroup:Refresh();
		animationGroup:Play();
]]

local lib = LibStub:NewLibrary("LibRecursiveAnim-1.0", "5");

if ( not lib ) then
	return;
end

-- array concatenation
function tConcat(...) 
    local t = {};
    for n = 1, select("#", ...) do
        local arg = select(n, ...);
		for _, v in ipairs(arg) do
			tinsert(t, v);
		end
    end
    return t;
end

-- check, if frame is parent's child
local function checkParent(frame, parent)
	local curParent = frame:GetParent();
	if ( curParent == parent ) then
		return true;
	elseif ( curParent == nil ) then
		return false;
	else
		return checkParent(curParent, parent);
	end
end

-- calculates origin offset
local function calcOrigin(animation, rootParent, point, offX, offY)

	local originX, originY;
	local frameX, frameY;
	
	-- calculating origin position
	if ( point == "TOPLEFT" ) then
		originX = rootParent:GetLeft();
		originY = rootParent:GetTop();
	elseif ( point == "LEFT" ) then
		originX = rootParent:GetLeft();
		originY = select(2, rootParent:GetCenter());
	elseif ( point == "BOTTOMLEFT" ) then
		originX = rootParent:GetLeft();
		originY = rootParent:GetBottom();
	elseif ( point == "BOTTOM" ) then
		originX = rootParent:GetCenter();
		originY = rootParent:GetBottom();
	elseif ( point == "BOTTOMRIGHT" ) then
		originX = rootParent:GetRight();
		originY = rootParent:GetBottom();
	elseif ( point == "RIGHT" ) then
		originX = rootParent:GetRight();
		originY = select(2, rootParent:GetCenter());
	elseif ( point == "TOPRIGHT" ) then
		originX = rootParent:GetRight();
		originY = rootParent:GetTop();
	elseif ( point == "TOP" ) then
		originX = rootParent:GetCenter();
		originY = rootParent:GetTop();
	else
		originX, originY = rootParent:GetCenter()
	end
	local rootScale = rootParent:GetEffectiveScale();
	originX = originX * rootScale + offX;
	originY = originY * rootScale + offY;
	
	-- calculating frame position
	local frame = animation:GetRegionParent();
	local frameScale = frame:GetEffectiveScale();
	frameX, frameY = frame:GetCenter();
	frameX = frameX * frameScale;
	frameY = frameY * frameScale;

	return "CENTER", originX - frameX, originY - frameY;
end

-- list of methods that should be called on each child AnimationGroup
local ANIM_GROUP_DIR_METHOD_ALL = {
    'Finish',
    'Pause',
    'Play',
    'SetInitialOffset',
    'SetLooping',
    'Stop',
};

-- list of methods that should be called on only one child AnimationGroup
local ANIM_GROUP_DIR_METHOD_SINGLE = {
    'GetAnimations',
    'GetDuration',
    'GetInitialOffset',
	'GetLoopState',
    'GetLooping',
    'GetMaxOrder',
    'GetProgress',
    'IsDone',
    'IsPaused',
    'IsPendingFinish',
    'IsPlaying',
};

-- list of methods, which calls should be logged
local ANIM_GROUP_LOG_METHODS = {
    'SetInitialOffset',
    'SetLooping',
};

-- list of methods that should be called on each child Animation
local ANIM_DIR_METHOD_ALL = {
    ['common'] = {
        'Pause',
        'Play',
        'SetDuration',
        'SetEndDelay',
        'SetMaxFramerate',
        'SetOrder',
        'SetSmoothProgress',
        'SetSmoothing',
        'SetStartDelay',
        'Stop',
    },

    ['Scale'] = {
        'SetScale',
    },

    ['Alpha'] = {
        'SetChange',
    },

    ['Rotation'] = {
        'SetDegrees',
        'SetRadians',
    },

    ['Translation'] = {
        'SetOffset',
    },

    ['Path'] = {
        'SetCurve',
    },
};

-- list of methods that should be called on only one child Animation
local ANIM_DIR_METHOD_SINGLE = {
    ['common'] = {
        'GetDuration',
        'GetElapsed',
        'GetEndDelay',
        'GetMaxFramerate',
        'GetOrder',
        'GetProgress',
        'GetProgressWithDelay',
        'GetSmoothProgress',
        'GetSmoothing',
        'GetStartDelay',
        'IsDelaying',
        'IsDone',
        'IsPaused',
        'IsPlaying',
        'IsStopped',
    },

    ['Scale'] = {
        'GetScale',
    },

    ['Alpha'] = {
        'GetChange',
    },

    ['Rotation'] = {
        'GetDegrees',
        'GetRadians',
    },

    ['Translation'] = {
        'GetOffset',
    },

    ['Path'] = {
        'GetCurve',
        'GetMaxOrder',
    },
    
};

-- list of methods, which calls should be logged
local ANIM_LOG_METHODS = {
	'SetDuration',
	'SetEndDelay',
	'SetMaxFramerate',
	'SetOrder',
	'SetSmoothProgress',
	'SetSmoothing',
	'SetStartDelay',

	-- Scale
	'SetScale',

    -- Alpha
	'SetChange',

    -- Rotation
	'SetDegrees',
	'SetRadians',

    -- Translation
	'SetOffset',

    -- Path
	'SetCurve',
};

local ANIM_METHODS = {
	['Alpha'] = tConcat(ANIM_DIR_METHOD_ALL['common']
		, ANIM_DIR_METHOD_ALL['Alpha']
		, ANIM_DIR_METHOD_SINGLE['common']
		, ANIM_DIR_METHOD_SINGLE['Alpha']),
	['Scale'] = tConcat(ANIM_DIR_METHOD_ALL['common']
		, ANIM_DIR_METHOD_ALL['Scale']
		, ANIM_DIR_METHOD_SINGLE['common']
		, ANIM_DIR_METHOD_SINGLE['Scale']),
	['Rotation'] = tConcat(ANIM_DIR_METHOD_ALL['common']
		, ANIM_DIR_METHOD_ALL['Rotation']
		, ANIM_DIR_METHOD_SINGLE['common']
		, ANIM_DIR_METHOD_SINGLE['Rotation']),
	['Translation'] = tConcat(ANIM_DIR_METHOD_ALL['common']
		, ANIM_DIR_METHOD_ALL['Translation']
		, ANIM_DIR_METHOD_SINGLE['common']
		, ANIM_DIR_METHOD_SINGLE['Translation']),
	['Path'] = tConcat(ANIM_DIR_METHOD_ALL['common']
		, ANIM_DIR_METHOD_ALL['Path']
		, ANIM_DIR_METHOD_SINGLE['common']
		, ANIM_DIR_METHOD_SINGLE['Path']),
}

local ANIM_GROUP_METHODS = tConcat(ANIM_GROUP_DIR_METHOD_ALL, ANIM_GROUP_DIR_METHOD_SINGLE);

--[[
    metapoint constructor
		parent           - parent metaanimation
		subpointIterator - function, that returns iterator over managed points
		setupTable       - table, in which logged method calls will be saved
		
	called by RecursiveAnimation:CreateControlPoint()
]]
local function CreateRecursiveControlPoint(name, order, parent, subpointIterator, setupTable)

	local _offsetX, _offsetY, _order, _parent = 0, 0, order, parent;

	-- object that will be returned
	local cp = {};
	
	if ( name ) then
		_G[name] = cp;
	end

	function cp:GetOrder()
		if ( _order ) then
			return _order;
		end
		for subpoint in subpointIterator() do
			return subpoint:GetOrder();
		end
	end

	function cp:GetOffset()
		return _offsetX, _offsetY;
	end

	function cp:SetOffset(offsetX, offsetY)
		_offsetX, _offsetY = offsetX, offsetY;
		for subpoint in subpointIterator() do
			subpoint:SetOffset(offsetX, offsetY);
		end
		if ( setupTable["SetOffset"] ) then
			setupTable["SetOffset"][1] = offsetX;
			setupTable["SetOffset"][2] = offsetY;
		else
			setupTable["SetOffset"] = {offsetX, offsetY};
		end
	end

	function cp:SetOrder(order)
		_order = order;
		for subpoint in subpointIterator() do
			subpoint:SetOrder(order);
		end
		if ( setupTable["SetOrder"] ) then
			setupTable["SetOrder"][1] = order;
		else
			setupTable["SetOrder"] = {order};
		end
	end

	function cp:GetName()
		return name;
	end

	function cp:GetObjectType()
		return "RecursiveControlPoint";
	end

	function cp:GetParent()
		return _parent;
	end

	function cp:IsObjectType(type)
		return (type == "RecursiveControlPoint");
	end

	return cp;
end

--[[
    metaanimation constructor
		parent          - parent metagroup
		subanimIterator - function, that returns iterator over managed animations
		getFirstAnim    - function, that returns one of managed animations (each time the same)
		rootParent      - root animated frame
		setupTable      - table in which saving info about called methods
		
	called by RecursiveAnimationGroup:CreateAnimation()
]]
local function CreateRecursiveAnimation(animationType, name, inheritsFrom, parent, subanimIterator, getFirstAnim, rootParent, setupTable)

	-- PRIVATE:

	local _custOriginPoint, _custOriginX, _custOriginY = "CENTER", 0, 0;

	local _metaControlPoints;

	-- hash: animation - control points array
	local _controlPointsByAnim;

    --[[
        array that stores info about created metapoints
          _controlPointInfos
            |-(table) - table index is metaanimation id, if 
               |-name (string|nil) - global name
               |-inheritsFrom (string|nil) - template
               |-order (number) - initial ordr value
               |-setup (table) - hash: method name - args array
    --]]
	local _controlPointInfos;

	local attachCP;

	-- object that will be returned
	local a = {};

	if ( name ) then
		_G[name] = a;
	end
	
	
	-- PUBLIC:
	
	-- generating method calls redirects
	for _, method in ipairs(ANIM_METHODS[animationType]) do
		local isLogged = tContains(ANIM_LOG_METHODS, method);
		local all = tContains(ANIM_DIR_METHOD_ALL['common'], method)
			or tContains(ANIM_DIR_METHOD_ALL[animationType], method);
			
		a[method] = function(self, ...)
			if ( isLogged ) then
				if ( setupTable[method] ) then
					wipe(setupTable[method]);
					for i = 1, select('#', ...) do
						setupTable[method][i] = select(i, ...);
					end
				else
					setupTable[method] = {...};
				end
			end
			if ( all ) then
				for animation in subanimIterator() do
					animation[method](animation, ...);
				end
			else
				local animation = getFirstAnim();
				return animation[method](animation, ...);
			end
		end
	end

	-- part of methods implemented here
	function a:GetName()
		return name;
	end

	function a:GetParent()
		return parent;
	end

	function a:GetRegionParent()
		return rootParent;
	end

	function a:GetObjectType()
		return "RecursiveAnimation";
	end

	function a:IsObjectType(type)
		return (type == "RecursiveAnimation");
	end

	function a:GetScript(scriptType, firstOnly)
		assert( firstOnly == nil or type(firstOnly) == "boolean"
			, "RecursiveAnimation:GetScript() Bad argument #3 (boolean expected, got " .. type(firstOnly) .. ")");

		local animation = getFirstAnim();

		if ( firstOnly ) then
			return animation:GetScript(scriptType);
		else
			for ianimation in subanimIterator() do
				if ( ianimation ~= animation ) then
					return ianimation:GetScript(scriptType);
				end
			end
			return animation:GetScript(scriptType);
		end
	end

	function a:HasScript(scriptType, checkFirstOnly)
		assert( firstOnly == nil or type(checkFirstOnly) == "boolean"
			, "RecursiveAnimation:HasScript() Bad argument #3 (boolean expected, got " .. type(checkFirstOnly) .. ")");

		local animation = getFirstAnim();

		if ( checkFirstOnly ) then
			return animation:HasScript(scriptType);
		else
			for ianimation in subanimIterator() do
				if ( ianimation ~= animation ) then
					return ianimation:HasScript(scriptType);
				end
			end
			return animation:HasScript(scriptType);
		end
	end

	function a:HookScript(scriptType, handler, firstOnly)
		assert( firstOnly == nil or type(firstOnly) == "boolean"
			, "RecursiveAnimation:HookScript() Bad argument #3 (boolean expected, got " .. type(firstOnly) .. ")");

		if ( firstOnly ) then
			local animation = getFirstAnim();
			animation:HookScript(scriptType, handler);
		else
			for animation in subanimIterator() do
				animation:HookScript(scriptType, handler);
			end
		end
	end

	function a:SetScript(scriptType, handler, firstOnly)
		assert( firstOnly == nil or type(firstOnly) == "boolean"
			, "RecursiveAnimation:SetScript() Bad argument #3 (boolean expected, got " .. type(firstOnly) .. ")");

		if ( firstOnly ) then

			local animation = getFirstAnim();
			animation:SetScript(scriptType, handler);
			for ianimation in subanimIterator() do
				if ( ianimation ~= animation ) then
					ianimation:SetScript(scriptType, nil);
				end
			end

		else
			for animation in subanimIterator() do
				animation:SetScript(scriptType, handler);
			end
		end
	end

	-- methods dependent on animation type
	if ( animationType == "Scale" or animationType == "Rotation" ) then

		function a:SetOrigin(point, offX, offY)
			assert( type(point) == "string" and
				( point == "CENTER" or
				point == "TOP" or
				point == "TOPRIGHT" or
				point == "TOPLEFT" or
				point == "LEFT" or
				point == "RIGHT" or
				point == "BOTTOM" or
				point == "BOTTOMRIGHT" or
				point == "BOTTOMLEFT" )
				, "RecursiveAnimation:SetOrigin() Bad argument #1 (point string expected, got " .. type(point) .. ")");
			assert( type(offX) == "number"
				, "RecursiveAnimation:SetOrigin() Bad argument #2 (number expected, got " .. type(offX) .. ")");
			assert( type(offY) == "number"
				, "RecursiveAnimation:SetOrigin() Bad argument #3 (number expected, got " .. type(offY) .. ")");
			
			_custOriginPoint, _custOriginX, _custOriginY = point, offX, offY;
			for animation in subanimIterator() do
				animation:SetOrigin(calcOrigin(animation, rootParent, point, offX, offY));
			end
		end

		function a:GetOrigin()
			return _custOriginPoint, _custOriginX, _custOriginY;
		end
		
	elseif ( animationType == "Path" ) then

		_metaControlPoints = {};

		_controlPointsByAnim = {};

		_controlPointInfos = {};

		-- attaches control points to animation
		attachCP = function(animation, cpId)
			local cpInfo = _controlPointInfos[cpId];
			local controlPoint = animation:CreateControlPoint(nil, cpInfo.inheritsFrom, cpInfo.order);

			-- calling logged methods
			for method, args in pairs(cpInfo.setup) do
				controlPoint[method](controlPoint, unpack(args));
			end
			
			_controlPointsByAnim[animation][cpId] = controlPoint;
		end

		function a:CreateControlPoint(name, inheritsFrom, order)

			assert( (name == nil) or type(name) == "string"
				, "RecursiveAnimation:CreateControlPoint() Bad argument #1 (string expected, got " .. type(name) .. ")");
			assert( (inheritsFrom == nil) or type(inheritsFrom) == "string"
				, "RecursiveAnimation:CreateControlPoint() Bad argument #2 (string expected, got " .. type(inheritsFrom) .. ")");
			assert( (order == nil) or type(order) == "number"
				, "RecursiveAnimation:CreateControlPoint() Bad argument #3 (number expected, got " .. type(order) .. ")");

			local _id = #_metaControlPoints + 1;
		
			local cpInfo = {
				['name']  = name,
				['inheritsFrom'] = inheritsFrom,
				['order'] = order,
				['setup'] = {},
			};
			tinsert(_controlPointInfos, cpInfo);
			
			-- returns iterator over control points managed by this metapoint
			local subpointIterator = function()
				local next = subanimIterator();
				return function()
					local anim = next();
					return anim and _controlPointsByAnim[anim][_id];
				end
			end

			local metaControlPoint = CreateRecursiveControlPoint(name
				, order
				, self
				, subpointIterator
				, cpInfo.setup);
			tinsert(_metaControlPoints, metaControlPoint);

			-- attaching control point to each animation
			for animation in subanimIterator() do
				attachCP(animation, _id);
			end
			
			return metaControlPoint;
		end

		function a:GetControlPoints()
			local res = {}
			for _, metaPoint in ipairs(_metaControlPoints) do
				tinsert(res, metaPoint);
			end
			return unpack(res);
		end
	end

	function a:Refresh()
		if ( animationType == "Path" ) then
		
			local toDeleteList = {};

			-- animations list, got from metagroup
			local actualAnimList = {};
			for animation in subanimIterator() do
				tinsert(actualAnimList, animation);
			end

			-- checking if there are detached animations
			for anim, points in pairs(_controlPointsByAnim) do
				local detached = not tContains(actualAnimList, anim);

				if ( detached ) then
					tinsert(toDeleteList, anim);
				end
			end

			for _, anim in ipairs(toDeleteList) do
				_controlPointsByAnim[anim] = nil;
			end

			-- checking if there are attached animations
			for _, anim in ipairs(actualAnimList) do
				local isNew = true;
				for animation in pairs(_controlPointsByAnim) do
					if ( animation == anim ) then
						isNew = false;
						break;
					end
				end

				-- if yes - attaching points
				if ( isNew ) then	
					_controlPointsByAnim[anim] = {};
					for id, _ in ipairs(_metaControlPoints) do
						attachCP(anim, id);
					end
				end
			end

		end

	end

	return a;
end


--- Creates recursiveAnimationGroup object. It have the same methods as AnimationGroup. For more info see
-- @param parent Frame that will be animated. (Frame)
-- @param name Optional, global identifier. (string)
-- @param inheritsFrom Optional, template name. (string)
-- @return RecursiveAnimationGroup object. (table)
function lib:CreateRecursiveAnimationGroup(parent, name, inheritsFrom)

	-- checking, if parent is a Region instance
	assert( type(parent) == "table"
		, "Bad argument #1 to CreateRecursiveAnimationGroup() (Region expected, got " .. type(parent) .. ")");
	assert( type(parent.IsObjectType) == "function"
		, "Bad argument #1 to CreateRecursiveAnimationGroup() (Region expected, but object got isn't an UIObject instance)");
	assert( parent:IsObjectType("Region")
		, "Bad argument #1 to CreateRecursiveAnimationGroup() (Region expected, got " .. parent:GetObjectType() .. ")");
	
	assert( (name == nil) or type(name) == "string"
		, "Bad argument #2 to CreateRecursiveAnimationGroup() (string expected, got " .. type(name) .. ")");
	assert( (inheritsFrom == nil) or type(inheritsFrom) == "string"
		, "Bad argument #3 to CreateRecursiveAnimationGroup() (string expected, got " .. type(inheritsFrom) .. ")");

	-- object that will be returned
    local g = {};

    if ( name ) then
        _G[name] = g;
    end
		
    -- PRIVATE:

    local _parent, _name, _inheritsFrom = parent, name, inheritsFrom;
	
	-- hashs: method name - args array
	local _groupSetup = {};

	-- metaanimations array
	local _metaAnims = {};

    -- hash: group - animations array
	local _animGroups = {};

	-- returns iterator over animations with given id
	local function animIter(id)
		local next, t, i = pairs(_animGroups);
		return function()
			local newI, animArray = next(t, i);
			
			-- skipping gap elements
			while ( animArray == 0 ) do
				newI, animArray = next(t, newI);
			end
			i = newI;
			
			return animArray and animArray[id];
		end
	end

	-- root animationGroup
	local _firstGroup = _parent:CreateAnimationGroup(_name, _inheritsFrom);

    -- hash: frame - has attached animations
	local _equipedChildren = {};

    --[[
        Array that stores info about created metaanimations
          _animInfos
            |-(table) - table index is metaanimation id, if 
               |-type (string) - animation type
               |-name (string|nil) - global name
               |-inheritsFrom (string|nil) - template
               |-setup (table) - hash: method name - args array
    --]]
    local _animInfos = {};

    -- function that attaches animation to group
    local function attachAnimation(animGroup, animId)
		local animInfo = _animInfos[animId];
        local animation = animGroup:CreateAnimation(animInfo.type, nil, animInfo.inheritsFrom);
		
        -- if necessary - setting origin
		if ( animInfo.type == "Scale" or animInfo.type == "Rotation" ) then
			animation:SetOrigin(calcOrigin(animation, _parent, "CENTER", 0, 0));
		end

		-- calling logged methods
        for method, args in pairs(animInfo.setup) do
			animation[method](animation, unpack(args));
		end
        
        _animGroups[animGroup][animId] = animation;
    end

    -- function that equips frame with animations
    local function equipChild(frame)		
        local animGroup = frame:CreateAnimationGroup(nil, _inheritsFrom);
        _animGroups[animGroup] = {};
		
		for method, args in pairs(_groupSetup) do
			animGroup[method](animGroup, unpack(args));
		end
       
        for animId, animInfo in ipairs(_animInfos) do
			
			-- we don't need attach alpha animation to child
			if ( animInfo.type ~= "Alpha" ) then
				attachAnimation(animGroup, animId);
			end
		end
    end
	
	-- function, that equips frame and his children with animations
	local function equipChildren(frame)
		local children = {frame:GetChildren()};

		for _, child in ipairs(children) do

			if ( not _equipedChildren[child] ) then
                equipChild(child);
				_equipedChildren[child] = true;
			end

            equipChildren(child);

		end
	end
	
	_animGroups[_firstGroup] = {};
    equipChildren(_parent);

	
    -- PUBLIC:	

    function g:CreateAnimation(animationType, name, inheritsFrom)
		assert( type(animationType) == "string"
			, "RecursiveAnimationGroup:CreateAnimation() Bad argument #1 (string expected, got " .. type(animationType) .. ")");
		assert( animationType == "Alpha"
			or animationType == "Scale"
			or animationType == "Path"
			or animationType == "Translation"
			or animationType == "Rotation"
			, "RecursiveAnimationGroup:CreateAnimation() Bad argument #1 (animationType string expected)");

		assert( (name == nil) or type(name) == "string"
			, "RecursiveAnimationGroup:CreateAnimation() Bad argument #2 (string expected, got " .. type(name) .. ")");
		assert( (inheritsFrom == nil) or type(inheritsFrom) == "string"
			, "RecursiveAnimationGroup:CreateAnimation() Bad argument #3 (string expected, got " .. type(inheritsFrom) .. ")");

		local _id = #_animInfos+1;
		
		local animInfo = {
			['type']  = animationType,
			['name']  = name,
			['inheritsFrom'] = inheritsFrom,
			['setup'] = {},
		};
		tinsert(_animInfos, animInfo);

		-- returns iterator over managed animations
		local subanimIterator = function()
			return animIter(_id);
		end

		-- returns animation that belongs to root frame
		local getFirstAnim = function()
			return _animGroups[_firstGroup][_id];
		end
		
		-- creating metaanimation
		local metaAnim = CreateRecursiveAnimation(animationType
		    , name
			, inheritsFrom
			, self
			, subanimIterator
			, getFirstAnim
			, _parent
			, animInfo.setup);
		tinsert(_metaAnims, metaAnim);    
		
		-- alpha animation applies on all subframes by default, so we don't attach it to children
		if ( animationType == "Alpha" ) then
			attachAnimation(_firstGroup, _id);
			for animGroup, animArray in pairs(_animGroups) do
				if ( animGroup ~= _firstGroup ) then
					animArray[_id] = 0;
				end
			end
		else
			for animGroup in pairs(_animGroups) do
				attachAnimation(animGroup, _id);
			end
		end

		metaAnim:Refresh();

		return metaAnim;
	end

	-- generating metagroup methods
	for _, method in ipairs(ANIM_GROUP_METHODS) do
		local all = tContains(ANIM_GROUP_DIR_METHOD_ALL, method);
		local isLogged = tContains(ANIM_GROUP_LOG_METHODS, method);
		g[method] = function(self, ...)
			if ( isLogged ) then
				if ( _groupSetup[method] ) then
					wipe(_groupSetup[method]);
					for i = 1, select('#', ...) do
						_groupSetup[method][i] = select(i, ...);
					end
				else
					_groupSetup[method] = {...};
				end
			end
			if ( all ) then
				for group in pairs(_animGroups) do
					group[method](group, ...);
				end
			else
				local group = _firstGroup;
				return group[method](group, ...);
			end
		end
	end

	function g:GetName()
		return _name;
	end

	function g:GetObjectType()
		return "RecursiveAnimationGroup";
	end

	function g:GetParent()
		return _parent;
	end

	function g:IsObjectType(type)
		return type == "RecursiveAnimationGroup";
	end

	function g:GetScript(scriptType, firstOnly)
		assert( firstOnly == nil or type(firstOnly) == "boolean"
			, "RecursiveAnimationGroup:GetScript() Bad argument #3 (boolean expected, got " .. type(firstOnly) .. ")");

		if ( firstOnly ) then
			return _firstGroup:GetScript(scriptType);
		else
			for animGroup in pairs(_animGroups) do
				if ( animGroup ~= _firstGroup ) then
					return animGroup:GetScript(scriptType);
				end
			end
			return _firstGroup:GetScript(scriptType);
		end
	end

	function g:HasScript(scriptType, checkFirstOnly)
		assert( checkFirstOnly == nil or type(checkFirstOnly) == "boolean"
			, "RecursiveAnimationGroup:HasScript() Bad argument #3 (boolean expected, got " .. type(checkFirstOnly) .. ")");

		if ( checkFirstOnly ) then
			return _firstGroup:HasScript(scriptType);
		else
			for animGroup in pairs(_animGroups) do
				if ( animGroup ~= _firstGroup ) then
					return animGroup:HasScript(scriptType);
				end
			end
			return _firstGroup:HasScript(scriptType);
		end
	end

	function g:HookScript(scriptType, handler, firstOnly)
		assert( firstOnly == nil or type(firstOnly) == "boolean"
			, "RecursiveAnimationGroup:HookScript() Bad argument #3 (boolean expected, got " .. type(firstOnly) .. ")");

		if ( firstOnly ) then
			_firstGroup:HookScript(scriptType, handler);
		else
			for animGroup in pairs(_animGroups) do
				animGroup:HookScript(scriptType, handler);
			end
		end
	end

	function g:SetScript(scriptType, handler, firstOnly)
		assert( firstOnly == nil or type(firstOnly) == "boolean"
			, "RecursiveAnimationGroup:SetScript() Bad argument #3 (boolean expected, got " .. type(firstOnly) .. ")");

		if ( firstOnly ) then

			_firstGroup:SetScript(scriptType, handler);
			for animGroup in pairs(_animGroups) do
				if ( animGroup ~= _firstGroup ) then
					animGroup:SetScript(scriptType, nil);
				end
			end

		else
			for animGroup in pairs(_animGroups) do
				animGroup:SetScript(scriptType, handler);
			end
		end
	end

    
    -- refreshes animations attachement
	function g:Refresh()
		equipChildren(_parent);

		-- checking if there are detached subframes
		for child in pairs(_equipedChildren) do
			if ( not checkParent(child, _parent) ) then
				local removedAnimGroups = {child:GetAnimationGroups()};

				-- deleting group that belongs to detached frame
				for animGroup, anims in pairs(_animGroups) do
					if ( tContains(removedAnimGroups, animGroup) ) then
						_animGroups[animGroup] = nil;
						break;
					end
				end
			end
		end

		-- refresh origins
		for animId, animInfo in ipairs(_animInfos) do
			if ( animInfo.type == "Scale" or animInfo.type == "Rotation" ) then
				for animation in animIter(animId) do
					animation:SetOrigin(calcOrigin(animation, _parent, _metaAnims[animId]:GetOrigin()));
				end
			end
		end

		for _, metaAnim in ipairs(_metaAnims) do
			metaAnim:Refresh();
		end
	end
	
	return g;
end